from matplotlib.colors import LinearSegmentedColormap
import matplotlib.pyplot as plt
import numpy as np
from PIL import Image

class ImageUtils:

    @staticmethod
    def colourise_heatmaps(heatmaps, image_width=224, image_height=224):

        hm_colours = [(1, 0, 0), (0, 1, 0), (0, 0, 1), (1, 1, 0), (0, 1, 1)]
        start_colour = (0, 0, 0)
        
        fused_heatmap_np = None
        
        for i, heatmap in enumerate(heatmaps):
            
            map_colours = [start_colour, hm_colours[i]]

            cmap_tp = LinearSegmentedColormap.from_list("Custom", map_colours, N=256)
            
            if len(heatmaps) == 1:
                cmap_tp = plt.get_cmap('jet')
            
            heatmap_image = Image.fromarray(np.uint8(heatmap*255), 'L').resize((image_width,image_height), resample=Image.BICUBIC)
            heatmap_np = cmap_tp(np.array(heatmap_image))[:, :, :3]
            
            if i == 0:
                fused_heatmap_np = heatmap_np
            else:
                fused_heatmap_np += heatmap_np

        fused_heatmap_np/=np.max(fused_heatmap_np)
                
        fused_heatmap_image = Image.fromarray(np.uint8((fused_heatmap_np * 255)), "RGB")
        
        return fused_heatmap_image
    
    @staticmethod
    def super_imposed_image(heatmaps, original_image, image_width=224, image_height=224, alpha=0.8):
        
        original_image = original_image.resize((image_width, image_height))
        original_image = Image.blend(original_image, heatmaps.convert("RGB"), alpha=alpha)

        return original_image


import gc
import torch
from torch.nn import functional as F
from ..utils import GradUtils
import queue


class FMGCam:

    
    def __init__(self, model, layer_name, device=None):
        self.model = model
        self.layer_name = layer_name

        if device is None:
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        else:
            self.device = device

        self.model.to(self.device)
        self.model.eval()

        self.my_queue = queue.Queue()
    
    # def storeInQueue(f, self=self):
    #     def wrapper(*args):
    #         self.my_queue.put(f(*args))
    #     return wrapper

    # @storeInQueue
    def __call__(self, img_tensor, class_count=3, enhance=True, class_rank_index=None):

        img_tensor = img_tensor.to(self.device)
        preds, sorted_pred_indices, grad_list, act_list = GradUtils.get_model_pred_with_grads(
            model=self.model, 
            img_tensor=img_tensor, 
            last_conv_layer=self.layer_name,
            class_count=class_count)

        heatmaps = GradUtils.weight_activations(act_list, grad_list)

        if class_rank_index is None:

            # Concatenation of activation maps based on top n classes
            heatmaps = torch.cat(heatmaps)

            # Filter the heatmap based on the maximum weighted activation along the channel axis
            hm_mask_indices = heatmaps.argmax(dim=0).unsqueeze(0)

            hm_3d_mask = torch.cat([hm_mask_indices for _ in range(heatmaps.size()[0])])

            hm_3d_mask = torch.cat(
                [(hm_3d_mask[index] == (torch.ones_like(hm_3d_mask[index])*index)).unsqueeze(0) for index in range(heatmaps.size()[0])]
            ).long()

            heatmaps *= hm_3d_mask

        else:

            heatmaps = heatmaps[class_rank_index]

        
        # L2 Normalisation of the heatmap soften the differences
        if enhance:
            heatmaps = F.normalize(heatmaps, p=2, dim=1)

        # relu on top of the heatmap
        heatmaps = F.relu(heatmaps)
        
        # Min-max normalization of the heatmap
        heatmaps = (heatmaps - torch.min(heatmaps))/(torch.max(heatmaps) - torch.min(heatmaps))
        
        # Process the generated heatmaps
        heatmaps = heatmaps.detach().cpu().numpy()

        gc.collect()
        self.my_queue.put((preds, sorted_pred_indices, heatmaps))

        return preds, sorted_pred_indices, heatmaps


import threading
import torch
from PIL import Image

import os
import sys

# sys.path.append(os.path.join(os.path.dirname(__file__), "."))

from .services import FMGCam
from .utils import ImageUtils

class XAIInferenceEngine:
    def __init__(
        self,
        model: torch.nn.Module,
        last_conv_layer: torch.nn.Module,
        device: torch.device = None,
        num_workers: int = 1,
    ):
        self.model = model
        self.num_workers = num_workers
        self.last_conv_layer = last_conv_layer

        if device is None:
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        else:
            self.device = device

        self.model.to(self.device)

        # Set the model to evaluation mode
        self.model.eval()

        self.fm_g_cam_generator = FMGCam(self.model, self.last_conv_layer, self.device)

    def predict(
        self,
        img: Image.Image,
        img_tensor: torch.Tensor,
        class_count: int = 3,
        class_rank_index: int = None,
        enhance: bool = True,
        alpha: float = 0.8,
        image_width: int = 224,
        image_height: int = 224,
    ):
        # Increase the stack size to prevent stack overflow  # TODO: Try to remove this
        threading.stack_size(1000000)
        fm_g_cam_generator_thread = threading.Thread(
            target=self.fm_g_cam_generator,
            kwargs={
                "img_tensor": img_tensor,
                "class_count": class_count,
                "enhance": enhance,
                "class_rank_index": class_rank_index,
            },
        )
        fm_g_cam_generator_thread.start()
        preds, sorted_pred_indices, heatmaps = self.fm_g_cam_generator.my_queue.get()

        heatmaps = ImageUtils.colourise_heatmaps(heatmaps)

        super_imp_img = ImageUtils.super_imposed_image(
            heatmaps,
            img,
            alpha=alpha,
            image_width=image_width,
            image_height=image_height,
        )

        return preds, sorted_pred_indices, super_imp_img, heatmaps



from PIL import Image
import torch
from torch.nn import functional as F
from torchvision.transforms.functional import to_pil_image, pil_to_tensor

import requests

import os
import sys
sys.path.append(os.path.join(os.path.dirname(__file__), "../"))

def run_test():
    
    print("[INFO]: Running test.py")

    print("[INFO]: Testing XAIInferenceEngine...")

    print("[INFO]: Importing Libraries...")
    from xai_inference_engine import XAIInferenceEngine
    from torchvision.models import resnet50, ResNet50_Weights

    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    print("[INFO]: Device: {}".format(device))

    print("[INFO]: Loading Model...")
    # Model
    model = resnet50(weights=ResNet50_Weights.IMAGENET1K_V2).to(device)
    weights = ResNet50_Weights.DEFAULT
    preprocess = weights.transforms()

    # Model config
    # Set model to eval mode
    model.eval()
    last_conv_layer = model.layer4[2].conv3
    class_count = 5
    class_list = weights.meta["categories"]
    img_h = 224

    print("[INFO]: Image Preprocessing...")
    # Image Preprocessing
    url = "https://raw.githubusercontent.com/utkuozbulak/pytorch-cnn-visualizations/master/input_images/cat_dog.png"
    r = requests.get(url, allow_redirects=True)
    open("dog-and-cat-cover.jpg", "wb").write(r.content)
    img = Image.open("dog-and-cat-cover.jpg")
    img = img.resize((img_h, img_h), resample=Image.BICUBIC)
    img_tensor = preprocess(img).to(device)


    print("[INFO]: Creating XAIInferenceEngine...")
    xai_inferencer = XAIInferenceEngine(
        model=model,
        last_conv_layer=last_conv_layer,
        device=device,
    )

    print("[INFO]: Running XAIInferenceEngine.predict()...")
    preds, sorted_pred_indices, super_imp_img, heatmaps = xai_inferencer.predict(
        img=img,
        img_tensor=img_tensor,
    )

    print("[INFO]: Saving Results...")
    super_imp_img.save("super_imp_img.jpg")
    heatmaps.save("heatmaps.jpg")

    print("[INFO]: Displaying Results...")
    print("        Predictions: {}".format(preds.shape))
    print("        Sorted Prediction Indices: {}".format(sorted_pred_indices.cpu().numpy()[:10]))
    print("        Heatmaps shape: {}".format(heatmaps))
    print("        Super Imposed Image: {}".format(super_imp_img))


if __name__ == "__main__":
    run_test()


from setuptools import setup, find_packages

NAME = 'xai_inference_engine'
AUTHOR = 'Ravidu Suien Rammuni Silva'
AUTHOR_EMAIL = 'ravidus.ac@gmail.com'
VERSION = '0.1.1' 
DESCRIPTION = 'A pip package for XAI Inferencing'
LONG_DESCRIPTION = 'TODO'

setup(
    name=NAME,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    version=VERSION,
    description=DESCRIPTION,
    packages=find_packages(
        where='src',
        include=['xai_inference_engine', 'xai_inference_engine.*']
    ),
    package_dir={'': 'src'},
    install_requires=[
        'colorama==0.4.6',
        'contourpy==1.1.1',
        'cycler==0.12.1',
        'filelock==3.12.4',
        'fonttools==4.43.1',
        'fsspec==2023.10.0',
        'Jinja2==3.1.2',
        'kiwisolver==1.4.5',
        'MarkupSafe==2.1.3',
        'matplotlib==3.8.0',
        'mpmath==1.3.0',
        'networkx==3.2',
        'numpy==1.26.1',
        'packaging==23.2',
        'Pillow==10.1.0',
        'pyparsing==3.1.1',
        'pyproject_hooks==1.0.0',
        'python-dateutil==2.8.2',
        'six==1.16.0',
        'sympy==1.12',
        'tomli==2.0.1',
        'torch==2.1.0',
        'typing_extensions==4.8.0',
    ],
    extras_require={
        'interactive': ['jupyter'],
    },
    # setup_requires=['pytest-runner', 'flake8'],
    # tests_require=['pytest'],
    # entry_points={
    #     'console_scripts': ['my-command=exampleproject.example:main'] # TODO: Command shortcuts
    # }
    # package_data={'': ['*.json']} # TODO: additional data files
)


import torch
from torch.nn import functional as F

class GradUtils:

    @staticmethod
    def get_model_pred_with_grads(model, img_tensor, last_conv_layer, class_count):
        # TODO: Search for the last convolutional layer
        
        grad_list = []
        act_list = []
        
        for train_param in model.parameters():
            train_param.requires_grad = True
            
        gradients = None
        activations = None

        def hook_backward(module, grad_input, grad_output):
            nonlocal gradients
            gradients = grad_output

        def hook_forward(module, args, output):
            nonlocal activations
            activations = output
            
            
        hook_backward = last_conv_layer.register_full_backward_hook(hook_backward, prepend=False)
        hook_forward = last_conv_layer.register_forward_hook(hook_forward, prepend=False)
        
        model.eval()
        
        preds =  model(img_tensor.unsqueeze(0))
        
        # Sort prediction indices
        sorted_pred_indices = torch.argsort(preds, dim=1, descending=True).squeeze(0)
        
        # Iterate through the top prediction indices
        for rank in range(class_count):
            preds[:, sorted_pred_indices[rank]].backward(retain_graph=True)
            grad_list.append(gradients)
            act_list.append(activations)
        
        hook_backward.remove()
        hook_forward.remove()
        
        for train_param in model.parameters():
            train_param.requires_grad = False
        
        return preds.squeeze().detach().cpu().numpy(), sorted_pred_indices, grad_list, act_list
    

    @staticmethod
    def weight_activations(activation_list, gradient_list):
        heatmaps = []

        for index, activations in enumerate(activation_list):
            gradients = gradient_list[index]
            
            avg_pooled_gradients = torch.mean(
                gradients[0], # Size [1, 1024, 7, 7] TODO - make this support other models
                dim=[0, 2, 3]
            )

            # Weighting acitvation features (channels) using its related calculated Gradient
            for i in range(activations.size()[1]):
                activations[:, i, :, :] *= avg_pooled_gradients[i]

            # average the channels of the activations
            heatmap = torch.mean(activations, dim=1).squeeze()
                    
            heatmaps.append(heatmap.unsqueeze(0).detach().cpu())

        avg_pooled_gradients = torch.mean(
            gradients[0], # Size [1, 1024, 7, 7] TODO - make this support other models
            dim=[0, 2, 3]
        )

        return heatmaps